/*****************************************************************************
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: phm2rs.c,v 1.3 2000/04/29 23:24:56 kevin Exp $
**  $Log: phm2rs.c,v $
**  Revision 1.3  2000/04/29 23:24:56  kevin
**  *** empty log message ***
**
**  Revision 1.2  2000/04/28 13:50:45  kevin
**  Removed Makefile Makefile.in that are automatically generated by autoconf
**
**  Revision 1.1.1.1  2000/04/28 13:02:44  kevin
**  Initial CVS import for first public release
**
**
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/
/* FILE
 *   phm2rs.c		Generate raysum projections from phantom object
 *
 * HISTORY
 *   1984 -- Final DOS version  
 *   1999 -- First UNIX version
 */

#include "ct.h"

#define O_PHANTOM    1
#define O_DESC       2
#define O_NRAY       3
#define O_ROTANGLE   4
#define O_TRACE      5
#define O_VERBOSE    6
#define O_HELP       7
#define O_PICFILE    8
#define O_DEBUG     10
#define O_VERSION   11

static struct option my_options[] = 
{
  {"phantom", 1, 0, O_PHANTOM},
  {"picfile", 1, 0, O_PICFILE},
  {"desc", 1, 0, O_DESC},
  {"nray", 1, 0, O_NRAY},
  {"rotangle", 1, 0, O_ROTANGLE},
  {"trace", 1, 0, O_TRACE},
  {"verbose", 0, 0, O_VERBOSE},
  {"help", 0, 0, O_HELP},
  {"debug", 0, 0, O_DEBUG},
  {"version", 0, 0, O_VERSION},
  {0, 0, 0, 0}
};

void 
usage (const char *program)
{
#ifdef MPI_CT
if (mpi_ct.my_rank == 0)
  {
#endif  
  fprintf(stdout,"usage: %s outfile ndet nview [--phantom phantom-name] [--picfile filename] [OPTIONS]\n", kbasename(program));
  fprintf(stdout,"Calculate raysums (projections) through phantom object, either\n");
  fprintf(stdout,"a predefined --phantom or a --picfile\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"     outfile      Name of output file for raysums\n");
  fprintf(stdout,"     ndet         Number of detectors\n");
  fprintf(stdout,"     nview        Number of rotated views\n");
  fprintf(stdout,"     --phantom    Phantom to use for projection\n");
  fprintf(stdout,"        herman    Herman head phantom\n");
  fprintf(stdout,"        rowland   Rowland head phantom\n");
  fprintf(stdout,"        browland  Bordered Rowland head phantom\n");
  fprintf(stdout,"        unitpulse Unit pulse phantom\n");
  fprintf(stdout,"     --picfile    Get Phantom from picture file\n");
  fprintf(stdout,"     --desc       Description of raysum\n");
  fprintf(stdout,"     --nray       Number of rays per detector (default = 1)\n");
  fprintf(stdout,"     --rotangle   Degrees to rotate view through, multiple of PI (default = 1)\n");
  fprintf(stdout,"     --trace      Trace level to use\n");
  fprintf(stdout,"        none      No tracing (default)\n");
  fprintf(stdout,"        text      Trace text level\n");
  fprintf(stdout,"        pic       Trace picture\n");
  fprintf(stdout,"        rays      Trace rays\n");
  fprintf(stdout,"        plot      Trace plot\n");
  fprintf(stdout,"        clipping  Trace clipping\n");
  fprintf(stdout,"     --verbose    Verbose mode\n");
  fprintf(stdout,"     --debug      Debug mode\n");
  fprintf(stdout,"     --version    Print version\n");
  fprintf(stdout,"     --help       Print this help message\n");
#ifdef MPI_CT
  }
  MPI_Abort(mpi_ct.comm, 1);
#endif
  exit(1);
}


void mpi_gather_rs (RAYSUM *rs_global, RAYSUM *rs_local, const int opt_debug);

int 
main (const int argc, char *const argv[])
{
  DETECTOR *det;
  PICTURE *pic = NULL;
  RAYSUM *rs_global;
  char str[256], *opt_outfile, opt_desc[MAXREMARK+1];
  char opt_picfilename[256];
  char *endptr, *endstr;
  int opt_ndet, opt_nview;
  int opt_nray = 1;
  int opt_trace = 0, opt_picnum = -1;
  int opt_verbose = 0;
  int opt_debug = 0;
  double opt_rotangle = 1;
  double time_start, time_end;
#ifdef MPI_CT
  RAYSUM *rs_local;
  int mpi_argc = argc;
  char **mpi_argv = (char **) argv;
  double mpi_t1, mpi_t2, mpi_t, mpi_t_g;

  MPI_Init(&mpi_argc, &mpi_argv);
  MPI_Comm_dup (MPI_COMM_WORLD, &mpi_ct.comm);
  MPI_Comm_size(mpi_ct.comm, &mpi_ct.nproc);
  MPI_Comm_rank(mpi_ct.comm, &mpi_ct.my_rank);

  if (mpi_ct.nproc > MPI_MAX_PROCESS) {
    sys_error(ERR_FATAL, "Number of mpi processes (%d) exceeds max processes (%d)",
	      mpi_ct.nproc, MPI_MAX_PROCESS);
    exit(1);
  }
#endif

#ifdef MPI_CT
  time_start = MPI_Wtime();
#else
  time_start = td_current_sec();
#endif
  
#ifdef MPI_CT
  if (mpi_ct.my_rank == 0) {
#endif
    strcpy(opt_desc, "");
    strcpy(opt_picfilename, "");
    while (1) {
      int c = getopt_long(argc, argv, "", my_options, NULL);
      char *endptr = NULL;
      char *endstr;
      
      if (c == -1)
	break;
      
      switch (c) {
      case O_PHANTOM:
	opt_picnum = opt_set_picture(optarg, argv[0]);
	pic = create_pic(opt_picnum);
	break;
      case O_PICFILE:
#ifdef MPI_CT
	if (mpi_ct.my_rank == 0)
	  fprintf(stderr, "Can not read picture from file in MPI mode\n");
	exit(1);
#endif
	strncpy(opt_picfilename, optarg, sizeof(opt_picfilename));
	pic = create_pic_from_file(opt_picfilename);
	break;
      case O_VERBOSE:
	opt_verbose = 1;
	break;
      case O_DEBUG:
	opt_debug = 1;
	break;
	break;
      case O_TRACE:
	opt_trace = opt_set_trace(optarg, argv[0]);
	break;
      case O_DESC:
	strncpy(opt_desc, optarg, sizeof(opt_desc));
	break;
      case O_ROTANGLE:
	opt_rotangle = strtod(optarg, &endptr);
	endstr = optarg + strlen(optarg);
	if (endptr != endstr) {
	  fprintf(stderr,"Error setting --rotangle to %s\n", optarg);
	  usage(argv[0]);
	  exit(1);
	}
	break;
      case O_NRAY:
	opt_nray = strtol(optarg, &endptr, 10);
	endstr = optarg + strlen(optarg);
	if (endptr != endstr) {
	  fprintf(stderr,"Error setting --nray to %s\n", optarg);
	  usage(argv[0]);
	  exit(1);
	}
	break;
        case O_VERSION:
#ifdef VERSION
	  fprintf(stdout, "Version %s\n", VERSION);
#else
	  fprintf(stderr, "Unknown version number");
#endif
	  exit(0);
      case O_HELP:
      case '?':
	usage(argv[0]);
	exit(0);
      default:
	usage(argv[0]);
	exit(1);
      }
    }
  
    if (pic == NULL) {
      fprintf(stderr, "No phantom defined\n");
      usage(argv[0]);
      exit(1);
    }
    if (optind + 3 != argc) {
      usage(argv[0]);
      exit(1);
    }
    opt_outfile = argv[optind];
    opt_ndet = strtol(argv[optind+1], &endptr, 10);
    endstr = argv[optind+1] + strlen(argv[optind+1]);
    if (endptr != endstr) {
      fprintf(stderr,"Error setting --ndet to %s\n", argv[optind+1]);
      usage(argv[0]);
      exit(1);
    }
    opt_nview = strtol(argv[optind+2], &endptr, 10);
    endstr = argv[optind+2] + strlen(argv[optind+2]);
    if (endptr != endstr) {
      fprintf(stderr,"Error setting --nview to %s\n", argv[optind+2]);
      usage(argv[0]);
      exit(1);
    }

    snprintf(str, sizeof(str), 
	     "Raysum_Collect: NDet=%d, Nview=%d, NRay=%d, RotAngle=%.2f, ", 
	     opt_ndet, opt_nview, opt_nray, opt_rotangle);
    if (opt_picfilename[0]) {
      strncat(str, "Phantom=", sizeof(str));
      strncat(str, opt_picfilename, sizeof(str));
    } else if (opt_picnum != -1) {
      strncat(str, "Phantom=", sizeof(str));
      strncat(str, name_of_picture(opt_picnum), sizeof(str));
    }
    if (opt_desc[0]) {
      strncat(str, ": ", sizeof(str));
      strncat(str, opt_desc, sizeof(str));
    }
    strncpy(opt_desc, str, sizeof(opt_desc));
#ifdef MPI_CT
  }
#endif

#ifdef MPI_CT
  MPI_Barrier(mpi_ct.comm);
  MPI_Bcast(&opt_rotangle, 1, MPI_DOUBLE, 0, mpi_ct.comm);
  MPI_Bcast(&opt_nview, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_ndet, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_nray, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_picnum, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_verbose, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_debug, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_trace, 1, MPI_INT, 0, mpi_ct.comm);

  if (mpi_ct.my_rank > 0 && opt_picnum > 0)
    pic = create_pic(opt_picnum);
#endif

  opt_rotangle *= PI;
  det = detect_create (pic, opt_ndet, opt_nview, opt_nray, opt_rotangle);

#ifdef MPI_CT
  mpi_ct_calc_work_units(opt_nview);
  if (mpi_ct.my_rank == 0) {
    rs_global = raysum_create_from_det (opt_outfile, det);
  }
  
  rs_local = raysum_create_from_det (NULL, det);
  rs_local->nview = mpi_ct.local_work_units[mpi_ct.my_rank];
  if (opt_debug)
    printf("rs_local->nview = %d (process %d)\n", rs_local->nview, mpi_ct.my_rank);

  if (opt_verbose)
    mpi_t1 = MPI_Wtime();
  raysum_collect (rs_local, det, pic, mpi_ct.start_work_unit[mpi_ct.my_rank], opt_trace, FALSE);
  if (opt_verbose) {
    mpi_t2 = MPI_Wtime();
    mpi_t = mpi_t2 - mpi_t1;
    MPI_Reduce(&mpi_t, &mpi_t_g, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_ct.comm);
    if (mpi_ct.my_rank == 0)
      printf("Time to collect rs = %f secs, Max = %f secs\n", mpi_t, mpi_t_g);
  }

  if (opt_verbose)
    mpi_t1 = MPI_Wtime();
  mpi_gather_rs(rs_global, rs_local, opt_debug);
  if (opt_verbose) {
    mpi_t2 = MPI_Wtime();
    mpi_t = mpi_t2 - mpi_t1;
    MPI_Reduce(&mpi_t, &mpi_t_g, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_ct.comm);
    if (mpi_ct.my_rank == 0)
      printf("Time to gather rs = %f secs, Max = %f secs\n", mpi_t, mpi_t_g);
  }
#else
  rs_global = raysum_create_from_det (opt_outfile, det);
  raysum_collect (rs_global, det, pic, 0, opt_trace, FALSE);
#endif
  
#ifdef MPI_CT
  time_end = MPI_Wtime();
  if (mpi_ct.my_rank == 0) {
    rs_global->calctime = time_end - time_start;
    strncpy (rs_global->remark, opt_desc, sizeof(rs_global->remark));
    if (opt_verbose) {
      fprintf(stdout, "Remark: %s\n", rs_global->remark);
      fprintf(stdout, "Time active = %.2f secs\n", rs_global->calctime);
    }
  }
#else
  time_end = td_current_sec();
  rs_global->calctime = time_end - time_start;
  strncpy (rs_global->remark, opt_desc, sizeof(rs_global->remark));
  if (opt_verbose) {
    fprintf(stdout, "Remark: %s\n", rs_global->remark);
    fprintf(stdout, "Time active = %.2f secs\n", rs_global->calctime);
  }
#endif

#ifdef MPI_CT
  if (mpi_ct.my_rank == 0) {
    raysum_write (rs_global);
    raysum_close (rs_global);
  }
#else  
  raysum_write (rs_global);
  raysum_close (rs_global);
#endif

  detect_free (det);

  if (opt_trace >= TRACE_PIC)
    {
      crt_set_cpos (1, 1);
      printf("Finished\n");
    }

  return (0);
}


#ifdef MPI_CT
void mpi_gather_rs (RAYSUM *rs_global, RAYSUM *rs_local, const int opt_debug)
{
  int iproc;
  int end_work_unit;
  int iw;

  end_work_unit = mpi_ct.local_work_units[mpi_ct.my_rank] - 1;
  for (iw = 0; iw <= end_work_unit; iw++) {
    MPI_Send(&rs_local->view[iw]->view_angle, 1, MPI_DOUBLE, 0, 0, mpi_ct.comm);
    MPI_Send(&rs_local->view[iw]->ndet, 1, MPI_INT, 0, 0, mpi_ct.comm);
    MPI_Send(rs_local->view[iw]->detval, rs_local->ndet, MPI_FLOAT, 0, 0, mpi_ct.comm);
  }

  if (mpi_ct.my_rank == 0) {
    for (iproc = 0; iproc < mpi_ct.nproc; iproc++) {
      end_work_unit = mpi_ct.start_work_unit[iproc] 
	+ mpi_ct.local_work_units[iproc] 
	- 1;

      if (opt_debug)
	fprintf(stdout, "Recv rs data from process %d\n", iproc);

      for (iw = mpi_ct.start_work_unit[iproc]; iw <= end_work_unit; iw++) {
	MPI_Status status;

	MPI_Recv(&rs_global->view[iw]->view_angle, 1, MPI_DOUBLE, iproc, 0, mpi_ct.comm, &status);
	MPI_Recv(&rs_global->view[iw]->ndet, 1, MPI_INT, iproc, 0, mpi_ct.comm, &status);
	MPI_Recv(rs_global->view[iw]->detval, rs_global->ndet, MPI_FLOAT, iproc, 0, mpi_ct.comm, &status);
      }
    }
  }

}
#endif
